/**
 * Add image upload dialog for Batchaction buttons.
 *
 * The `.productUiVariationuploadimagesUnity()` method adds functionality to open an image upload dialog.
 *
 *
 * ### Examples
 * Apply .productUiVariationuploadimagesUnity() to Button element.
 *
 * JavaScript:
 *
 *     de_epages('button')
 *         .productUiVariationuploadimagesUnity({
 *              items: function() {
 *                  var itemList = Array.from($firstContentList.find('input[name=ListObjectID]:checked')),
 *                      idList = itemList.map(function(checkbox) {
 *                           return checkbox.value;
 *                      }).join(',');
 *
 *                  return idList;
 *              },
 *              changeAction: 'JSONUploadVariationImage',
 *              viewAction: '#INPUT.ViewAction',
 *              objectId: #ObjectID
 *              #IF(#DEFINED(#WebInterfaceConf.MAXCONTENTLENGTH)), maxFileSize: #WebInterfaceConf.MAXCONTENTLENGTH#ENDIF
 *          });
 *
 * HTML:
 *
 *     <button>Click Me!</button>
 *
 *
 * @class jQuery.ui.productUiVariationuploadimagesUnity
 *
 * @uses de_epages
 * @uses ep
 * @uses ep.fn.busy
 * @uses ep.ui.dialog
 * @uses jquery.ui.widget
 * @since 7.2.0
 */

/**
 * @cfg {String} changeAction Name of the Changeaction for submitting the image
 * @cfg {String} viewAction Name of the ViewAction for submitting the image
 * @cfg {String} objectId objectId of the master product
 * @cfg {String} maxFileSize max size for uploading images
 * @cfg {Function} items Function returning an Array of ObjectIds of the target variations
 */

/*
 * @copyright       © Copyright 2006-2010, epages GmbH, All Rights Reserved.
 *
 * @module          de_epages/product/ui/variationuploadimages-unity
 */
define('de_epages/product/ui/variationuploadimages-unity', [
    'jquery',
    'ep',
    'de_epages',
    '$tmpl!de_epages/product/ui/variationuploadimages-unity',
    '$dict!de_epages/product/ui/variationuploadimages-unity',

    "jquery/ui/widget",
    "ep/fn/busy",
    'ep/ui/dialog'
], function ($, ep, de_epages, tmplUploadForm, dict) {

    /*
     * @dictionary      ep.dict
     *
     * @translations    {Save}
     *                  {Cancel}
     *                  {or}
     *                  {UploadImage}
     *                  {InputError}
     *                  {UnknownError}
     *                  {UploaderError}
     *                  {FileScaleError}
     *                  {MimetypeNotAcceptError}
     *                  {FileSizeLimitError}
     *
     *
     * @dictionary      de_epages.product.dictionary
     *
     * @translations    {UnityFileAssignmentError}
     *
     *
     * @dictionary      de_epages.product.ui.variationuploadimages
     *
     * @translation    {UnityUseMasterProductImage}
     */

    // translation object
    var translations = {
        Save: dict.translate('Save'),
        Cancel: dict.translate('Cancel'),
        Or: dict.translate('or'),

        UploadImage: dict.translate('UploadImage'),
        UnknownError: dict.translate('UnknownError'),
        InputError: dict.translate('InputError'),
        MimetypeNotAcceptError: dict.translate('MimetypeNotAcceptError'),
        FileSizeLimitError: dict.translate('FileSizeLimitError'),
        FILE_UPLOAD_ERROR: dict.translate('UploaderError'),
        FILE_SCALE_ERROR: dict.translate('FileScaleError'),
        IMAGE_ASSIGNMENT_ERROR: dict.translate('UnityFileAssignmentError'),

        UseMasterProductImage: dict.translate('UnityUseMasterProductImage')
    };

    $.widget('ui.productUiVariationuploadimagesUnity', {
    //  dndImage,
    //  formData,

        templateElements: {},

        options: {
            objectId: 1,
            changeAction: 'myDummyChangeAction',
            viewAction: '?',
            maxFileSize: 4096,
            mime: 'image/(gif|png|jpeg|bmp)',
            items: function() {
                return [];
            }
        },

        _create: function() {
            var self = this,
                elem = self.element,
                o = self.options,
                templateOptions = [{
                    tUploadImages: translations.UploadImage,
                    tUseMasterProductImage: translations.UseMasterProductImage,
                    tErrorHeadline: translations.InputError,
                    tOr: translations.Or
                }];

            // load template and push named tmpl elements to current instance
            self.dialogContent = tmplUploadForm(templateOptions);
            $.extend(self.templateElements, self.dialogContent.tmplItem('elements'));

            // register open dialog action to batch action button
            elem.on('click', function(evt) {
                self._openDialog();
            });

            self._openDialog();

            // event handler image uploader
            self.templateElements.useMasterProductImage.on('change', function (evt) {
                self._handleUseMaster(evt);
            });

            self.templateElements.fileInput.on('change', function(evt) {
                self._handleFile(evt);
            });

            // register DnD handling
            self.templateElements.dropzone.on('dragover', function(evt) {
                self._handleDragOver(evt);
            });

            self.templateElements.dropzone.on('drop', function(evt) {
                self._handleDrop(evt);
            });

            // register handler for drag and drop and file input
            // do not allow dropping on body
            $('body').on('dragover', function (e) {
                e.stopPropagation();
                e.preventDefault();
                e.originalEvent.dataTransfer.dropEffect = 'none';
            });
        },

        _openDialog: function() {
            var self = this,
                o = self.options,
                idString = o.items();

            self.templateElements.attributeIDs.val(idString);

            // open if dialog exists, otherwise create new dialog
            if (self.dialog) {
                self.dialog.uiDialog('open');

            } else {
                self.dialog = ep(self.dialogContent).uiDialog({
                    modal: true,
                    autoOpen: true,
                    width: 400,
                    title: translations.UploadImage,
                    close: function(){
                        self._resetPreviewImage();
                        self._resetForm();
                    },
                    buttons:{
                        Ok:     {
                            // type: 'submit',
                            click: function () {

                                self.dialog.parent().busy('show');

                                // build formData Object for submit
                                delete self.formData;
                                self.formData = new FormData();

                                // Add files if exist
                                if (self.templateElements.fileInput[0].files.length || self.dndImage) {
                                    self.formData.append("VariationImageUploaderOwnImage", self.dndImage || self.templateElements.fileInput[0].files[0]);
                                }

                                // set reset to master product image property if checked
                                if (self.templateElements.useMasterProductImage.prop('checked')) {
                                    self.formData.append("UseMasterProductImage", 1);
                                }

                                // add general attributes (changeAction, masterproduct-objectid, subproduct-objectids)
                                self.formData.append("AttributeIDs", self.templateElements.attributeIDs.val());
                                self.formData.append('ChangeAction', o.changeAction);
                                self.formData.append('ObjectID', o.objectId);

                                // send Ajax request
                                ep.ajax({
                                    url: '?',
                                    type: 'post',
                                    processData: false,
                                    contentType: false,
                                    data: self.formData,
                                    dataType: "json"
                                })
                                .done(function(data, textStatus, jqXHR) {
                                    if (data && data.Success) {
                                        // goto target location
                                        location.href='?ViewAction=' + o.viewAction + '&ObjectID=' + o.objectId;
                                    }
                                })
                                .fail(function(jqXHR, textStatus, errorThrown) {
                                    // remove busy and show errors
                                    self.dialog.parent().busy('hide');
                                    self._showErrors(errorThrown);
                                })
                                .always(function(data, textStatus, jqXHR) {});
                            },
                            text: translations.Save,
                            disabled: true
                        },
                        Cancel: {
                            click: function(){
                                ep(this).uiDialog('close');
                            },
                            text: translations.Cancel
                        }
                    }
                });
            }
        },

        // UseMaster-Checkbox
        _handleUseMaster: function(evt) {
            var self = this;

            self._hideErrors();
            self._checkSaveButton();

            // reset preview image
            if (!!self.templateElements.useMasterProductImage.prop('checked')) {
                self.dndImage = undefined;
                self._resetPreviewImage();
            }
        },

        // FileInput
        _handleFile: function(evt) {
            var self = this;

            // clear dnd image
            self.dndImage = undefined;

            // if file is selected handle file, otherwise reset preview
            if (evt.originalTarget.files.length) {
                self._fileReader(evt.originalTarget.files[0]);
            } else {
                self._resetPreviewImage();
                self._checkSaveButton();
            }
        },

        // DnD
        _handleDragOver: function(evt) {
            evt.stopPropagation();
            evt.preventDefault();

            evt.originalEvent.dataTransfer.dropEffect = 'copy';
        },

        // Drop-Handler
        _handleDrop: function(evt) {
            var self = this;

            evt.stopPropagation();
            evt.preventDefault();

            // reset file input, save dropped image
            self._resetPreviewImage();
            self.dndImage = evt.originalEvent.dataTransfer.files[0];
            self._fileReader(evt.originalEvent.dataTransfer.files[0]);
        },

        // FileReader
        _fileReader: function(file){
            var self = this,
                elem = self.element,
                o = self.options,
                reader = new FileReader();

            reader.onload = function(evt){
                var fileReader = this;

                // call handler to process file
                self._readerSuccessHandler(fileReader, file, evt);
            };

            reader.onloadstart = function (evt) {
                var errorList = [];

                // if a maximum file size error occurs
                if (Math.round(file.size/1000, 0) > self.options.maxFileSize) {
                    errorList.push({reason: 'FileSizeLimitError'});
                }

                // if a mime type error occurs
                if (!file.type.match(self.options.mime)) {
                    errorList.push({reason: 'MimetypeNotAcceptError'});
                }

                // if any error occurs cancel reading process and reset form otherwise hide any error
                if (errorList.length) {
                    reader.abort();

                    self._showErrors(errorList);
                    self._resetPreviewImage();
                    self._resetForm();
                    self._checkSaveButton();

                    return;
                } else {
                    self._hideErrors();
                }
            };

            reader.readAsDataURL(file);
        },

        _readerSuccessHandler: function(fileReader, file, evt) {
            var self = this,
                img = new Image();

            // image source contains content of FileReader
            img.src = fileReader.result;

            // load handler for image
            img.addEventListener('load', function(evt) {
                // display new image, hide icon
                self.templateElements.dropzone.addClass('ep-variationuploadimage-dropzone-handle');
                self.templateElements.dropzone.css('background-image', 'url(' + img.src + ')');
                self.templateElements.dropzoneIcon.hide();
                self.templateElements.useMasterProductImage.prop('checked', false);
                self._checkSaveButton();
            }, false);
        },

        _hideErrors: function() {
            var self = this;

            self.templateElements.errorBox.addClass('HideElement');
        },

        _showErrors: function(errorList) {
            var self = this,
                message;

            // empty error element, then push in all errors
            self.templateElements.errorMessage.html('');
            errorList.forEach(function(error) {
                message = translations[error.reason];
                self.templateElements.errorMessage.append('<li>' + (message ? message : translations.UnknownError) + '</li>');

            });

            // show error element
            self.templateElements.errorBox.removeClass('HideElement');
        },

        _checkSaveButton: function() {
            var self = this,
                imageUploaded = self.templateElements.fileInput.prop('files').length,
                imageDropped = self.dndImage,
                useMasterProductImage = self.templateElements.useMasterProductImage.prop('checked');

            // if either image is in file input, on dnd var or use master is checked, enable ok button
            self.dialog.uiDialog('option', 'buttons').Ok.prop('disabled', !(imageUploaded || useMasterProductImage || imageDropped));
        },

        _resetPreviewImage: function() {
            var self = this;

            self.templateElements.dropzone.removeClass('ep-variationuploadimage-dropzone-handle');
            self.templateElements.dropzone.css('background-image', '');
            self.templateElements.dropzoneIcon.show();
            self.templateElements.fileInput.val('');
        },

        _resetForm: function() {
            var self = this;

            // clear dnd image
            self.dndImage = undefined;

            // clear formData obj
            delete self.formData;

            // empty all inputs
            self.templateElements.attributeIDs.val('');
            self.templateElements.useMasterProductImage.prop('checked', false);
            self.templateElements.fileInput.val('');
        },

        destroy: function () {
            var self = this;

            self.dialog.uiDialog('destroy');
            self._superApply(arguments);
        }
    });

    return de_epages;
});
