/**
 * Create an ebay category controller and displays a structure to choose a category. The corresponding data are stored on server side.
 *
 * The `.ebayCategorycontroller()` method creates an ebay category controller.
 *
 * ### Examples
 * Create a drag and drop uploader and preload three images.

 * require([
 *     "jquery",
 *     "ep",
 *
 *     "de_epages/ebay/ui/ebaycategorycontroller",
 *     "$ready!"
 * ], function($, ep) {
 *     ep('.ep-ebay-controller').ebayCategorycontroller({
 *         option1: value1,
 *         option2: value2,
 *         option3: value3,
 *         option4: value4
 *     });
 * });
 *
 * HTML:
 *
 * <div class="ep-ebay-controller"></div>
 *
 * @class de_epages.ebay.ui.ebayCategorycontroller
 * @extends jQuery.widget
 *
 * @uses jQuery.dict
 * @uses jQuery.tmpl
 * @uses jQuery.i18n
 * @uses jQuery.ui.widget
 * @uses de_epages
 * @since 7.3.0
 */

/**
 * @cfg {Object} [controller] A json object with the initial category structure.
 */

/**
 * @cfg {String} [changeAction] ChangeAction for the ajax request to store category settings.
 */

/**
 * @cfg {String} [option3] An option with a value to configure the widget.
 */

/**
 * @cfg {String} [option4] An option with a value to configure the widget.
 */

/**
 * See `jQuery.ui.ebayCategorycontroller` for details.
 *
 * @param {Object} [controller] A json object with the initial category structure.
 * @param {String} [changeAction] ChangeAction for the ajax request to store category settings.
 * @param {Number} [objectId] ObjectId of the current Auction.
 * @param {Element} [busyTarget] Element to append busy screen to.
 * @param {Object} [dialogCtrl] Object to configure the category browser dialog.
 * @param {Object} [categorySelectors] Object to configure the different category selections.
 *
 * @method ebayCategorycontroller
 * @member jQuery
 *
 * @since 7.3.0
 */

/*
 * @copyright       © Copyright 2006-2010, epages GmbH, All Rights Reserved.
 *
 * @module          de_epages.ebay.ui.ebaycategorycontroller
 */

define( "de_epages/ebay/ui/ebaycategorycontroller", [
    "jquery",
    "ep",
    "de_epages",
    "util/storage",
    "$dict!../dictionary",
    "$tmpl!./ebaycategorycontroller",
    "$tmpl!./ebaycategorycontroller-parentlist",
    "$tmpl!./ebaycategorycontroller-breadcrump",
    "$tmpl!./ebaycategorycontroller-pathoutput",
    "$tmpl!./ebaycategorycontroller-error",

    "ep/fn/busy",
    "ep/ui/dialog",
    "jquery/tmpl"
], function($, ep, de_epages, storage, ebayDictionary, tmplEbayCategoryController, tmplEbayCategoryControllerParentList, tmplEbayCategoryControllerBreadcrump, tmplEbayCategoryControllerPathoutput, tmplEbayCategoryControllerError) {
    "use strict";

    /*
     * @dictionary      ep.dict
     *
     * @translation     {Apply}
     *                  {Error}
     *                  {Remove}
     *                  {UnityUnknownError}
     *
     * @dictionary      de_epages.ebay.dictionary
     *
     * @translation     {UnityAllCategories}
     *                  {UnityCategoryBrowserDialogTitle}
     *                  {UnitySelectedCategory}
     *                  {UnityNoCategorySelected}
     *                  {UnityUnkownCategoryNumber}
     *                  {UnitySelectByCategoryNumber}
     *                  {UnityCategoryLoadError}
     *                  {noShopEnvironment}
     *                  {noToken}
     *                  {EbayRequestFailed}
     */

    // translation object
    var translations = {
        allCategories: ebayDictionary.translate('UnityAllCategories'),
        categoryBrowserDialogTitle: ebayDictionary.translate('UnityCategoryBrowserDialogTitle'),
        selectedCategory: ebayDictionary.translate('UnitySelectedCategory'),
        selectByCategoryNumber: ebayDictionary.translate('UnitySelectByCategoryNumber'),
        noCategorySelected: ebayDictionary.translate('UnityNoCategorySelected'),
        unkownCategoryNumber: ebayDictionary.translate('UnityUnkownCategoryNumber'),
        categoryLoadError: ebayDictionary.translate('UnityCategoryLoadError'),
        EbayWarningNoEbayEnvironment: ebayDictionary.translate('noShopEnvironment'),
        NoEbayEnvironment: ebayDictionary.translate('noShopEnvironment'),
        EbayWarningNoToken: ebayDictionary.translate('noToken'),
        EbayRequestNotSuccessful: ebayDictionary.translate('EbayRequestFailed'),

        UnknownError: ebayDictionary.translate('UnityUnknownError'),
        error: ebayDictionary.translate('Error'),
        remove: ebayDictionary.translate('Remove'),
        apply: ebayDictionary.translate('Apply')
    };

    $.widget('ui.ebayCategorycontroller', {
    //  dom,
    //  parentList,
    //  dialog,

        /**
         * Options to be passed to the widgt on initialzation
         *
         * @cfg {Object} options
         * @cfg {Object} options.controller             controller object for rendering the categories
         * @cfg {String} [options.changeAction='dummy'] ChangeAction for getting category data
         * @cfg {Number} [options.objectId=0]           objectId of the current Auction
         * @cfg {Element} [options.busyTarget=Element]  element to append busy screen to
         * @cfg {Object} [options.dialogCtrl=Object]    Object to configure the category browser dialog
         * @cfg {Object} [categorySelectors=Object]     Object to configure the different category selections.
         */
        options: {
        //  controller,
        //  pathOutput,
        //  categorySelectors,

        //  dialogCtrl: {
        //    btnCat1: $('#cat1Button'),
        //    btnCat2: $('#cat2Button'),
        //    form: $categoryControllerElem.closest('form'),
        //    dialoClass: 'ep-ebay-categorybrowserdialog'
        //  },

            changeAction: 'dummy',
            objectId: 0,
            busyTarget: $('body')
        },

        /**
         * Magic jquery function.
         *
         * @method _create
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _create: function() {
            var self = this,
                elem = self.element,
                o = self.options,
                renderedTmpl,
                input;

            // if no controller object exists => leave widget
            if (!o.controller || !(o.categorySelectors && Array.isArray(o.categorySelectors) && o.categorySelectors.length>0)) {
                return;
            }
            // if no controller object exists => leave widget
            if (o.dialogCtrl && !self._isJQueryElement(o.dialogCtrl.dialogRoot)) {
                console.log('No dialog root');
                return;
            }
            // test busy element
            if (o.busyTarget) {
                o.busyTarget = self._isJQueryElement(o.busyTarget) ? o.busyTarget : elem;
            }

            // render view template for ebay categories
            renderedTmpl = tmplEbayCategoryController([{
                tCategoryPathLabel: translations.selectedCategory,
                tSelectByCategoryNumber: translations.selectByCategoryNumber
            }]);

            // // initialize object for category DOM elements
            self.dom = {};

            // // load template and push named tmpl elements to current instance
            $.extend(self.dom, renderedTmpl.tmplItem('elements'));
            // // append base template
            elem.append(renderedTmpl);

            // write category paths to DOM elements
            o.categorySelectors.forEach(function(item, index) {

                if (item.storedId !== undefined) {
                    item.currentSelectedLeaf = item.storedId;
                    self._setInitialCategoryPath(item.storedId, item, index);
                }

                item.pathOutput.on('click', 'span.fa-trash', function() {
                    self._deleteCategory(index);
                });
            });

            // set active category selector
            self.setActiveCategorySelector(0);

            self.dom.categoryInput.on('input', function(evt) {
                if (self.hTimer) {
                    clearTimeout(self.hTimer);
                    self.hTimer = undefined;
                }
                self.hTimer = setTimeout($.proxy(self, '_inputHandler'), 600);
            });

            // register events
            self.dom.parentCategory.on('click', 'li', self, self._handleListClick);
            self.dom.breadcrump.on('click', 'li', self, self._handleListClick);

            // register events for category select buttons
            if (o.dialogCtrl) {
                // register handle for first category button
                if (self._isJQueryElement(o.dialogCtrl.btnCat1)) {
                    o.dialogCtrl.btnCat1.on('click', function() {
                        self._handleCategoryButtonClick(0);
                    });
                }

                // register handle for second category button
                if (self._isJQueryElement(o.dialogCtrl.btnCat2)) {
                    o.dialogCtrl.btnCat2.on('click', function() {
                        self._handleCategoryButtonClick(1);
                    });
                }
            }
        },

        /**
         * Method to select the category (in epages two ebay categories can be stored).
         *
         * @method setActiveCategorySelector
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        setActiveCategorySelector: function(index) {
            var self = this,
                o = self.options;

            if (self.activeSelectorIndex !== index) {

                self.activeSelectorIndex = index;
                self.activeCategorySelector = o.categorySelectors[index];

                // render current category
                self._loadCategory(self.activeCategorySelector.currentId, function() {
                    if (self.activeCategorySelector.currentSelectedLeaf) {
                        self._setCategoryPath(self.activeCategorySelector.currentSelectedLeaf);
                        self.dom.categoryInput.val(o.controller[self.activeCategorySelector.currentSelectedLeaf].ExternalCategoryId);
                    } else {
                        // render path to dialog element
                        self.dom.categoryPath.empty();
                        self.dom.categoryInput.val('');
                    }
                });
            }
        },

        /**
         * Method to clicks on category select buttons.
         *
         * @method setActiveCategorySelector
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _handleCategoryButtonClick: function(index) {
            var self = this,
                o = self.options,
                initialResize = false;

            if (!self.dialog) {
                self._initDialog();
                initialResize = true;
            }

            self.setActiveCategorySelector(index);
            self.dialog.uiDialog('open');

            if (initialResize) {
                self._resizeDialog();
            }
        },

        /**
         * Method to create a dialog.
         *
         * @method _initDialog
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _initDialog: function() {
            var self = this,
                o = self.options,
                elem = self.element;

            self.dialog = o.dialogCtrl.dialogRoot.uiDialog({
                modal: true,
                autoOpen: false,
                dialogClass: o.dialogCtrl.dialogClass ? o.dialogCtrl.dialogClass : 'ep-ebay-categorybrowserdialog',
                appendTo: o.dialogCtrl.form ? o.dialogCtrl.form : $('body'),
                title: translations.categoryBrowserDialogTitle,
                buttons: [{
                    text: translations.apply,
                    click: function() {
                        self.applyData = true;
                        $(this).dialog("close");
                    }
                }],

                open: function() {
                    self.applyData = false;

                    self.initialState = {
                        leaf: self.activeCategorySelector.currentSelectedLeaf,
                        currentId: self.activeCategorySelector.currentId
                    }

                    self._hideErrors();
                    self._resizeDialog();
                },
                close: function() {
                    if (!self.applyData) {
                        self.activeCategorySelector.currentSelectedLeaf = self.initialState.leaf;
                        self.activeCategorySelector.currentId = self.initialState.currentId;

                        // render current category
                        self._loadCategory(self.activeCategorySelector.currentId, function() {
                            if (self.activeCategorySelector.currentSelectedLeaf) {
                                self._setCategoryPath(self.activeCategorySelector.currentSelectedLeaf);
                            } else {
                                // render path to dialog element
                                self.dom.categoryPath.empty();
                            }
                        });
                    }
                }
            });

            self.dialog.removeClass('HideElement');
        },

        _resizeDialog: function() {
            var self = this,
                o = self.options,
                dialogRoot = o.dialogCtrl.dialogRoot;

            if (self.dialog) {
                self.parentList.parentlist.height(dialogRoot.height() - dialogRoot.find('.ContentToggleAreaList').outerHeight(true) - self.breadcrumb.breadcrumblist.outerHeight(true) - self.dom.categoryPathWrapper.outerHeight(true) - self.dom.categoryNumberInputWrapper.outerHeight(true) - self.dom.errorMsg.outerHeight(true) - 20);
            }
        },

        /**
         * This method handles changes in the category input.
         *
         * @method _inputHandler
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _inputHandler: function() {
            var self = this,
                o = self.options;

                if (self.tooltip) {
                    self.tooltip.uiTooltip('hide');
                }

                ep.ajax({
                    type: 'POST',
                    dataType: 'json',
                    data:   {
                        ObjectID: o.objectId,
                        ChangeAction: 'JSONUnityGetInternalCategoryId',
                        ExternalCategoryID: self.dom.categoryInput.val()
                    }
                })
                .done(function(data) {
                    var ParentCategoryID;

                    if (data.Success) {
                        ParentCategoryID = (data.ParentCategoryID !== null) ? data.ParentCategoryID : 0;
                        self.activeCategorySelector.currentSelectedLeaf = data.EbayCategoryID;

                        self._loadCategory(ParentCategoryID, function() {
                            self._setCategoryPath(data.EbayCategoryID);
                            self.activeCategorySelector.catOutput.val(o.controller[data.EbayCategoryID].ExternalCategoryId);
                        });

                    } else {
                        // error handling
                        self._handleTooltip({
                            message: translations.unkownCategoryNumber
                        });
                    }
                })
                .fail(self._handleError);

            self.hTimer = undefined;
        },

        /**
         * This method maps the click on the list to the appropriate function.
         *
         * @method _handleListClick
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _handleListClick: function(evt) {
            var self = evt.data,
                o = self.options,
                $target = $(evt.currentTarget),
                categoryId = $target.data('ebay');

            // if clicked element is a folder
            if ($target.hasClass('ep-ebay-category-folder')) {
                self._loadCategory(categoryId);
            } else {
                // hide errors
                self._hideErrors();

                // add id of currently selected leaf
                self.activeCategorySelector.currentSelectedLeaf = categoryId;
                self._setCategoryPath(categoryId);

                if (self._isJQueryElement(self.activeCategorySelector.catOutput)) {
                    self.activeCategorySelector.catOutput.val(o.controller[categoryId].ExternalCategoryId);
                    self.activeCategorySelector.catOutput.trigger('change');
                }

                // check radio button
                $target.find('input[type=radio]').prop('checked', true);
            }

            self.dom.categoryInput.val(self.activeCategorySelector.currentSelectedLeaf ? o.controller[self.activeCategorySelector.currentSelectedLeaf].ExternalCategoryId : '');
        },

        /**
         * This method renders and sets the category path.
         *
         * @method _setCategoryPath
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _setInitialCategoryPath: function(categoryId, categorySelector, index) {
            var self = this,
                o = self.options,
                path,
                pathTmpl;

            if (self._isJQueryElement(categorySelector.pathOutput)) {
                path = categoryId
                    ? o.controller[categoryId].Path.map(function(item) {
                        return item.Name;
                    })
                    : [translations.noCategorySelected];

                // render parent list template
                pathTmpl = tmplEbayCategoryControllerPathoutput([{
                    path: path,
                    del: (categoryId === 0) ? false : true,
                    tDel: translations.remove
                }]);

                // render path to document element
                categorySelector.pathOutput.html(pathTmpl);
            }
        },

        /**
         * This method renders and sets the category path.
         *
         * @method _setCategoryPath
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _setCategoryPath: function(categoryId) {
            var self = this,
                o = self.options,
                path,
                pathTmpl;

            if (self._isJQueryElement(self.activeCategorySelector.pathOutput)) {
                path = o.controller[self.activeCategorySelector.currentId].Path.map(function(item) {
                    return item.Name;
                });
                path.push(o.controller[categoryId].Name);

                // render parent list template
                pathTmpl = tmplEbayCategoryControllerPathoutput([{
                    path: path,
                    del: (categoryId === 0) ? false : true,
                    tDel: translations.remove
                }]);

                // render path to document element
                self.activeCategorySelector.pathOutput.html(pathTmpl);

                // render parent list template
                pathTmpl = tmplEbayCategoryControllerPathoutput([{
                    path: path,
                    del: false,
                    tDel: translations.remove
                }]);
                // render path to dialog element
                self.dom.categoryPath.html(pathTmpl);
            }
        },

        /**
         * This method unselects the chosen category.
         *
         * @method _deleteCategory
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _deleteCategory: function(index) {
            var self = this,
                o = self.options;

            self._setInitialCategoryPath(0, o.categorySelectors[index], index);
            o.categorySelectors[index].currentId = 0;
            o.categorySelectors[index].currentSelectedLeaf = undefined;
            o.categorySelectors[index].catOutput.val('');

            self.activeSelectorIndex = undefined;
        },

        /**
         * This method loads a new category and appends it to the controller.
         *
         * @method _loadCategory
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _loadCategory: function(id, cb) {
            var self = this,
                o = self.options,
                busyElem = self.dialog ? self.dialog : o.busyTarget,
                localStorage;

            self._hideErrors();

            if (o.controller[id] && o.controller[id].updated) {
                self._renderParentCategoryList(id);
                // id for current category
                self.activeCategorySelector.currentId = id;

                if (cb && typeof cb === "function") {
                    cb();
                }
            } else {
                busyElem.busy('show');

                ep.ajax({
                    type: 'POST',
                    dataType: 'json',
                    data: {
                        ChangeAction: 'JSONUnityGetSubCategoryListById',
                        ObjectID: o.objectId,
                        ParentCategoryID: id
                    }
                })
                .done(function(json){
                    var innerCurrentId;

                    if (json.Success) {
                        // object for current category
                        self.activeCategorySelector.currentId = json.CurrentCategory.EbayCategoryID ? json.CurrentCategory.EbayCategoryID : 0;

                        // add response category to controller
                        o.controller[id] = json.CurrentCategory;
                        o.controller[id].updated = true;
                        o.controller[id].children = json.CategoryList.map(function(item) {
                            return item.EbayCategoryID;
                        });

                        // subcategory list for current category
                        json.CategoryList.forEach(function(category, index, array) {
                            innerCurrentId = category.EbayCategoryID;

                            if (!o.controller[innerCurrentId]) {
                                o.controller[innerCurrentId] = category;
                                o.controller[innerCurrentId].updated = false;

                                if (!o.controller[innerCurrentId].IsLeaf) {
                                    o.controller[innerCurrentId].children = [];
                                }
                            }
                        });


                        // if local storage object doesn't exist
                        if (!storage.localStorage('ebayStorage')) {
                            storage.localStorage('ebayStorage', {});
                        }

                        localStorage = storage.localStorage('ebayStorage');
                        // set local storage variable
                        localStorage.version = o.ebayCatVersion;
                        localStorage.controller = o.controller;
                        storage.localStorage('ebayStorage', localStorage);

                        self._renderParentCategoryList(id);
                        if (cb && typeof cb === "function") {
                            cb();
                        }
                    } else {
                        self._renderErrors(translations.categoryLoadError, json.doneErrors.map(function(error) {
                            return {
                                msg: translations[error.reason] ? translations[error.reason] : translations['UnknownError']
                            };
                        }));
                    }
                })
                .fail(self._handleError)
                .always(function() {
                    busyElem.busy('hide');
                });
            }
        },

        /**
         * This method renders the parent categories.
         *
         * @method _renderParentCategoryList
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _renderParentCategoryList: function(currentId) {
            var self = this,
                o = self.options,
                parentListTmpl;

            // render parent list template
            parentListTmpl = tmplEbayCategoryControllerParentList([{
                controller: o.controller,
                currentId: currentId,
                currentSelectedLeaf: self.activeCategorySelector.currentSelectedLeaf ? self.activeCategorySelector.currentSelectedLeaf : self.activeCategorySelector.storedId
            }]);

            // // initialize object for parent list
            self.parentList = {};

            // // load template and push named tmpl elements to current instance
            $.extend(self.parentList, parentListTmpl.tmplItem('elements'));
            self.parentList.parentlist = parentListTmpl;
            self.dom.parentCategory.html(parentListTmpl);
            self._renderBreadcrump(currentId);

            self._resizeDialog();
        },

        /**
         * This method renders the breadcrump.
         *
         * @method _renderParentCategoryList
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _renderBreadcrump: function(currentId) {
            var self = this,
                o = self.options,
                breadcrumpTmpl;

            // render parent list template
            breadcrumpTmpl = tmplEbayCategoryControllerBreadcrump([{
                path: [{
                    EbayCategoryID: 0,
                    Name: translations.allCategories
                }].concat(o.controller[currentId].Path)
            }]);

            // // initialize object for parent list
            self.breadcrumb = {};

            // // load template and push named tmpl elements to current instance
            $.extend(self.breadcrumb, breadcrumpTmpl.tmplItem('elements'));
            self.breadcrumb.breadcrumblist = breadcrumpTmpl;
            self.dom.breadcrump.html(breadcrumpTmpl);
        },

        /**
         * This method renders an error list.
         *
         * @method _renderErrors
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _renderErrors: function(title, errors) {
            var self = this,
                errorTmpl;

            // render parent list template
            errorTmpl = tmplEbayCategoryControllerError([{
                errors: errors,
                tTitle: title
            }]);

            // // load template and push named tmpl elements to current instance
            self.dom.errorMsg
                .html(errorTmpl)
                .removeClass('HideElement');

            // resize dialog content
            self._resizeDialog();
        },

        /**
         * This method empties and hides an error list.
         *
         * @method _hideErrors
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _hideErrors: function() {
            var self = this;

            // reset error element
            self.dom.errorMsg
                .empty()
                .addClass('HideElement');

            // resize dialog content
            self._resizeDialog();
        },

        /**
         * This method handles ajax failures.
         *
         * @method _handleError
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _handleError: function(data){
            if (data.status === 403) {
                location.href += "&SessionTimeout=1";
            } else {
                console.log("error object: ", data.responseJSON);
            }
        },

        /**
         * This method instantiates a tooltip.
         *
         * @method _initTooltip
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _initTooltip: function(options) {
            var context = options.context;

            this.tooltip = $('<div>')
                .uiTooltip({
                    interactive: false,
                    event: 'focus',
                    hideDelay: 1400,
                    context: context,
                    orientation: 'top',
                    hideDelay: 200,
                    oldhandling: true,
                    additionalClass: 'ep-formfield-tooltip ep-uiValidate-info'
                });
        },

        /**
         * This method displays a tooltip.
         *
         * @method _handleTooltip
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _handleTooltip: function(options) {
            var self = this,
                context = self.dom.categoryInput;

            // create a tooltip if necessary
            if (self.tooltip === undefined) {
                self._initTooltip({
                    context: context
                });
            }

            // update tooltip options
            self.tooltip.html(options.message);
            self.tooltip.uiTooltip('show');
        },

        /**
         * This method checks if an element is a jQuery object (returns true/false).
         *
         * @method _isJQueryElement
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        _isJQueryElement: function(element) {
            return (element && (typeof element === "object" && element[0] && element[0].tagName));
        },

        /**
         * This method removes the container on which the Uploader has been appended to.
         *
         * @method destroy
         * @member de_epages.ebay.ui.ebayCategorycontroller
         *
         * @since 7.3.0
         */
        destroy: function() {
            var self = this;

            self.elem.off('.ebayCategorycontroller');

            self._superApply(arguments);
        }

    });

    return de_epages;

});
